@@ -1,6 +1,6 @@
 /* Redis CLI (command line interface)
  *
- * Copyright (c) 2006-2009, Salvatore Sanfilippo <antirez at gmail dot com>
+ * Copyright (c) 2009-2010, Salvatore Sanfilippo <antirez at gmail dot com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -39,6 +39,7 @@
 #include "sds.h"
 #include "adlist.h"
 #include "zmalloc.h"
+#include "linenoise.h"
 
 #define REDIS_CMD_INLINE 1
 #define REDIS_CMD_BULK 2
@@ -51,6 +52,8 @@ static struct config {
     int hostport;
     long repeat;
     int dbnum;
+    int interactive;
+    char *auth;
 } config;
 
 struct redisCommand {
@@ -60,9 +63,12 @@ struct redisCommand {
 };
 
 static struct redisCommand cmdTable[] = {
+    {"auth",2,REDIS_CMD_INLINE},
     {"get",2,REDIS_CMD_INLINE},
     {"set",3,REDIS_CMD_BULK},
     {"setnx",3,REDIS_CMD_BULK},
+    {"append",3,REDIS_CMD_BULK},
+    {"substr",4,REDIS_CMD_INLINE},
     {"del",-2,REDIS_CMD_INLINE},
     {"exists",2,REDIS_CMD_INLINE},
     {"incr",2,REDIS_CMD_INLINE},
@@ -71,6 +77,8 @@ static struct redisCommand cmdTable[] = {
     {"lpush",3,REDIS_CMD_BULK},
     {"rpop",2,REDIS_CMD_INLINE},
     {"lpop",2,REDIS_CMD_INLINE},
+    {"brpop",-3,REDIS_CMD_INLINE},
+    {"blpop",-3,REDIS_CMD_INLINE},
     {"llen",2,REDIS_CMD_INLINE},
     {"lindex",3,REDIS_CMD_INLINE},
     {"lset",4,REDIS_CMD_BULK},
@@ -96,8 +104,13 @@ static struct redisCommand cmdTable[] = {
     {"zincrby",4,REDIS_CMD_BULK},
     {"zrem",3,REDIS_CMD_BULK},
     {"zremrangebyscore",4,REDIS_CMD_INLINE},
+    {"zmerge",-3,REDIS_CMD_INLINE},
+    {"zmergeweighed",-4,REDIS_CMD_INLINE},
     {"zrange",-4,REDIS_CMD_INLINE},
+    {"zrank",3,REDIS_CMD_BULK},
+    {"zrevrank",3,REDIS_CMD_BULK},
     {"zrangebyscore",-4,REDIS_CMD_INLINE},
+    {"zcount",4,REDIS_CMD_INLINE},
     {"zrevrange",-4,REDIS_CMD_INLINE},
     {"zcard",2,REDIS_CMD_INLINE},
     {"zscore",3,REDIS_CMD_BULK},
@@ -132,10 +145,24 @@ static struct redisCommand cmdTable[] = {
     {"debug",-2,REDIS_CMD_INLINE},
     {"mset",-3,REDIS_CMD_MULTIBULK},
     {"msetnx",-3,REDIS_CMD_MULTIBULK},
+    {"monitor",1,REDIS_CMD_INLINE},
+    {"multi",1,REDIS_CMD_INLINE},
+    {"exec",1,REDIS_CMD_INLINE},
+    {"discard",1,REDIS_CMD_INLINE},
+    {"hset",4,REDIS_CMD_MULTIBULK},
+    {"hget",3,REDIS_CMD_BULK},
+    {"hdel",3,REDIS_CMD_BULK},
+    {"hlen",2,REDIS_CMD_INLINE},
+    {"hkeys",2,REDIS_CMD_INLINE},
+    {"hvals",2,REDIS_CMD_INLINE},
+    {"hgetall",2,REDIS_CMD_INLINE},
+    {"hexists",3,REDIS_CMD_BULK},
+    {"config",-2,REDIS_CMD_BULK},
     {NULL,0,0}
 };
 
 static int cliReadReply(int fd);
+static void usage();
 
 static struct redisCommand *lookupCommand(char *name) {
     int j = 0;
@@ -148,14 +175,16 @@ static struct redisCommand *lookupCommand(char *name) {
 
 static int cliConnect(void) {
     char err[ANET_ERR_LEN];
-    int fd;
+    static int fd = ANET_ERR;
 
-    fd = anetTcpConnect(err,config.hostip,config.hostport);
     if (fd == ANET_ERR) {
-        fprintf(stderr,"Connect: %s\n",err);
-        return -1;
+        fd = anetTcpConnect(err,config.hostip,config.hostport);
+        if (fd == ANET_ERR) {
+            fprintf(stderr, "Could not connect to Redis at %s:%d: %s", config.hostip, config.hostport, err);
+            return -1;
+        }
+        anetTcpNoDelay(NULL,fd);
     }
-    anetTcpNoDelay(NULL,fd);
     return fd;
 }
 
@@ -185,6 +214,7 @@ static int cliReadSingleLineReply(int fd, int quiet) {
     if (reply == NULL) return 1;
     if (!quiet)
         printf("%s\n", reply);
+    sdsfree(reply);
     return 0;
 }
 
@@ -259,8 +289,7 @@ static int cliReadReply(int fd) {
     }
 }
 
-static int selectDb(int fd)
-{
+static int selectDb(int fd) {
     int retval;
     sds cmd;
     char type;
@@ -275,15 +304,15 @@ static int selectDb(int fd)
     if (type <= 0 || type != '+') return 1;
     retval = cliReadSingleLineReply(fd,1);
     if (retval) {
-        close(fd);
-	return retval;
+        return retval;
     }
     return 0;
 }
 
-static int cliSendCommand(int argc, char **argv) {
+static int cliSendCommand(int argc, char **argv, int repeat) {
     struct redisCommand *rc = lookupCommand(argv[0]);
     int fd, j, retval = 0;
+    int read_forever = 0;
     sds cmd;
 
     if (!rc) {
@@ -296,6 +325,7 @@ static int cliSendCommand(int argc, char **argv) {
             fprintf(stderr,"Wrong number of arguments for '%s'\n",rc->name);
             return 1;
     }
+    if (!strcasecmp(rc->name,"monitor")) read_forever = 1;
     if ((fd = cliConnect()) == -1) return 1;
 
     /* Select db number */
@@ -304,8 +334,8 @@ static int cliSendCommand(int argc, char **argv) {
         fprintf(stderr,"Error setting DB num\n");
         return 1;
     }
- 
-    while(config.repeat--) {
+
+    while(repeat--) {
         /* Build the command to send */
         cmd = sdsempty();
         if (rc->flags & REDIS_CMD_MULTIBULK) {
@@ -334,13 +364,16 @@ static int cliSendCommand(int argc, char **argv) {
         }
         anetWrite(fd,cmd,sdslen(cmd));
         sdsfree(cmd);
+
+        while (read_forever) {
+            cliReadSingleLineReply(fd,0);
+        }
+
         retval = cliReadReply(fd);
         if (retval) {
-            close(fd);
             return retval;
         }
     }
-    close(fd);
     return 0;
 }
 
@@ -349,7 +382,7 @@ static int parseOptions(int argc, char **argv) {
 
     for (i = 1; i < argc; i++) {
         int lastarg = i==argc-1;
-        
+
         if (!strcmp(argv[i],"-h") && !lastarg) {
             char *ip = zmalloc(32);
             if (anetResolve(NULL,argv[i+1],ip) == ANET_ERR) {
@@ -358,6 +391,8 @@ static int parseOptions(int argc, char **argv) {
             }
             config.hostip = ip;
             i++;
+        } else if (!strcmp(argv[i],"-h") && lastarg) {
+            usage();
         } else if (!strcmp(argv[i],"-p") && !lastarg) {
             config.hostport = atoi(argv[i+1]);
             i++;
@@ -367,6 +402,11 @@ static int parseOptions(int argc, char **argv) {
         } else if (!strcmp(argv[i],"-n") && !lastarg) {
             config.dbnum = atoi(argv[i+1]);
             i++;
+        } else if (!strcmp(argv[i],"-a") && !lastarg) {
+            config.auth = argv[i+1];
+            i++;
+        } else if (!strcmp(argv[i],"-i")) {
+            config.interactive = 1;
         } else {
             break;
         }
@@ -391,8 +431,59 @@ static sds readArgFromStdin(void) {
     return arg;
 }
 
+static void usage() {
+    fprintf(stderr, "usage: redis-cli [-h host] [-p port] [-a authpw] [-r repeat_times] [-n db_num] [-i] cmd arg1 arg2 arg3 ... argN\n");
+    fprintf(stderr, "usage: echo \"argN\" | redis-cli [-h host] [-a authpw] [-p port] [-r repeat_times] [-n db_num] cmd arg1 arg2 ... arg(N-1)\n");
+    fprintf(stderr, "\nIf a pipe from standard input is detected this data is used as last argument.\n\n");
+    fprintf(stderr, "example: cat /etc/passwd | redis-cli set my_passwd\n");
+    fprintf(stderr, "example: redis-cli get my_passwd\n");
+    fprintf(stderr, "example: redis-cli -r 100 lpush mylist x\n");
+    fprintf(stderr, "\nRun in interactive mode: redis-cli -i or just don't pass any command\n");
+    exit(1);
+}
+
+/* Turn the plain C strings into Sds strings */
+static char **convertToSds(int count, char** args) {
+  int j;
+  char **sds = zmalloc(sizeof(char*)*count+1);
+
+  for(j = 0; j < count; j++)
+    sds[j] = sdsnew(args[j]);
+
+  return sds;
+}
+
+static void repl() {
+    int size = 4096, max = size >> 1, argc;
+    char *line;
+    char **ap, *args[max];
+
+    while((line = linenoise("redis> ")) != NULL) {
+        if (line[0] != '\0') {
+          linenoiseHistoryAdd(line);
+          argc = 0;
+
+          for (ap = args; (*ap = strsep(&line, " \t")) != NULL;) {
+              if (**ap != '\0') {
+                  if (argc >= max) break;
+                  if (strcasecmp(*ap,"quit") == 0 || strcasecmp(*ap,"exit") == 0)
+                      exit(0);
+                  ap++;
+                  argc++;
+              }
+          }
+
+          cliSendCommand(argc, convertToSds(argc, args), 1);
+        }
+
+        free(line);
+    }
+
+    exit(0);
+}
+
 int main(int argc, char **argv) {
-    int firstarg, j;
+    int firstarg;
     char **argvcopy;
     struct redisCommand *rc;
 
@@ -400,34 +491,33 @@ int main(int argc, char **argv) {
     config.hostport = 6379;
     config.repeat = 1;
     config.dbnum = 0;
+    config.interactive = 0;
+    config.auth = NULL;
 
     firstarg = parseOptions(argc,argv);
     argc -= firstarg;
     argv += firstarg;
-    
-    /* Turn the plain C strings into Sds strings */
-    argvcopy = zmalloc(sizeof(char*)*argc+1);
-    for(j = 0; j < argc; j++)
-        argvcopy[j] = sdsnew(argv[j]);
 
-    if (argc < 1) {
-        fprintf(stderr, "usage: redis-cli [-h host] [-p port] [-r repeat_times] [-n db_num] cmd arg1 arg2 arg3 ... argN\n");
-        fprintf(stderr, "usage: echo \"argN\" | redis-cli [-h host] [-p port] [-r repeat_times] [-n db_num] cmd arg1 arg2 ... arg(N-1)\n");
-        fprintf(stderr, "\nIf a pipe from standard input is detected this data is used as last argument.\n\n");
-        fprintf(stderr, "example: cat /etc/passwd | redis-cli set my_passwd\n");
-        fprintf(stderr, "example: redis-cli get my_passwd\n");
-        fprintf(stderr, "example: redis-cli -r 100 lpush mylist x\n");
-        exit(1);
+    if (config.auth != NULL) {
+        char *authargv[2];
+
+        authargv[0] = "AUTH";
+        authargv[1] = config.auth;
+        cliSendCommand(2, convertToSds(2, authargv), 1);
     }
 
+    if (argc == 0 || config.interactive == 1) repl();
+
+    argvcopy = convertToSds(argc, argv);
+
     /* Read the last argument from stdandard input if needed */
     if ((rc = lookupCommand(argv[0])) != NULL) {
-        if (rc->arity > 0 && argc == rc->arity-1) {
-            sds lastarg = readArgFromStdin();
-            argvcopy[argc] = lastarg;
-            argc++;
-        }
+      if (rc->arity > 0 && argc == rc->arity-1) {
+        sds lastarg = readArgFromStdin();
+        argvcopy[argc] = lastarg;
+        argc++;
+      }
     }
 
-    return cliSendCommand(argc, argvcopy);
+    return cliSendCommand(argc, argvcopy, config.repeat);
 }
